@@ -46,296 +46,13 @@ OPENSSL_MSVC_PRAGMA(warning(pop))
 #endif
 
 #if !defined(OPENSSL_WINDOWS)
-using Socket = int;
-#define INVALID_SOCKET (-1)
-static int closesocket(int sock) { return close(sock); }
-static std::string LastSocketError() { return strerror(errno); }
+static const int kOpenReadOnlyBinary = O_RDONLY;
+static const int kOpenReadOnlyText = O_RDONLY;
 #else
-using Socket = SOCKET;
-static std::string LastSocketError() {
-  char buf[DECIMAL_SIZE(int) + 1];
-  snprintf(buf, sizeof(buf), "%d", WSAGetLastError());
-  return buf;
-}
+static const int kOpenReadOnlyBinary = _O_RDONLY | _O_BINARY;
+static const int kOpenReadOnlyText = O_RDONLY | _O_TEXT;
 #endif
 
-class OwnedSocket {
- public:
-  OwnedSocket() = default;
-  explicit OwnedSocket(Socket sock) : sock_(sock) {}
-  OwnedSocket(OwnedSocket &&other) { *this = std::move(other); }
-  ~OwnedSocket() { reset(); }
-  OwnedSocket &operator=(OwnedSocket &&other) {
-    reset(other.release());
-    return *this;
-  }
-
-  bool is_valid() const { return sock_ != INVALID_SOCKET; }
-  Socket get() const { return sock_; }
-  Socket release() {
-    Socket temp = std::move(sock_);
-    sock_ = INVALID_SOCKET;
-    return temp;
-  }
-
-  void reset(Socket sock = INVALID_SOCKET) {
-    if (is_valid()) {
-      closesocket(sock_);
-    }
-
-    sock_ = sock;
-  }
-
- private:
-  Socket sock_ = INVALID_SOCKET;
-};
-
-struct SockaddrStorage {
-  SockaddrStorage() : storage(), len(sizeof(storage)) {}
-
-  int family() const { return storage.ss_family; }
-
-  sockaddr *addr_mut() { return reinterpret_cast<sockaddr *>(&storage); }
-  const sockaddr *addr() const {
-    return reinterpret_cast<const sockaddr *>(&storage);
-  }
-
-  sockaddr_in ToIPv4() const {
-    if (family() != AF_INET || len != sizeof(sockaddr_in)) {
-      abort();
-    }
-    // These APIs were seemingly designed before C's strict aliasing rule, and
-    // C++'s strict union handling. Make a copy so the compiler does not read
-    // this as an aliasing violation.
-    sockaddr_in ret;
-    OPENSSL_memcpy(&ret, &storage, sizeof(ret));
-    return ret;
-  }
-
-  sockaddr_in6 ToIPv6() const {
-    if (family() != AF_INET6 || len != sizeof(sockaddr_in6)) {
-      abort();
-    }
-    // These APIs were seemingly designed before C's strict aliasing rule, and
-    // C++'s strict union handling. Make a copy so the compiler does not read
-    // this as an aliasing violation.
-    sockaddr_in6 ret;
-    OPENSSL_memcpy(&ret, &storage, sizeof(ret));
-    return ret;
-  }
-
-  sockaddr_storage storage;
-  socklen_t len;
-};
-
-static OwnedSocket Bind(int family, const sockaddr *addr, socklen_t addr_len) {
-  OwnedSocket sock(socket(family, SOCK_STREAM, 0));
-  if (!sock.is_valid()) {
-    return OwnedSocket();
-  }
-
-  if (bind(sock.get(), addr, addr_len) != 0) {
-    return OwnedSocket();
-  }
-
-  return sock;
-}
-
-static OwnedSocket ListenLoopback(int backlog) {
-  // Try binding to IPv6.
-  sockaddr_in6 sin6;
-  OPENSSL_memset(&sin6, 0, sizeof(sin6));
-  sin6.sin6_family = AF_INET6;
-  if (inet_pton(AF_INET6, "::1", &sin6.sin6_addr) != 1) {
-    return OwnedSocket();
-  }
-  OwnedSocket sock =
-      Bind(AF_INET6, reinterpret_cast<const sockaddr *>(&sin6), sizeof(sin6));
-  if (!sock.is_valid()) {
-    // Try binding to IPv4.
-    sockaddr_in sin;
-    OPENSSL_memset(&sin, 0, sizeof(sin));
-    sin.sin_family = AF_INET;
-    if (inet_pton(AF_INET, "127.0.0.1", &sin.sin_addr) != 1) {
-      return OwnedSocket();
-    }
-    sock = Bind(AF_INET, reinterpret_cast<const sockaddr *>(&sin), sizeof(sin));
-  }
-  if (!sock.is_valid()) {
-    return OwnedSocket();
-  }
-
-  if (listen(sock.get(), backlog) != 0) {
-    return OwnedSocket();
-  }
-
-  return sock;
-}
-
-static bool SocketSetNonBlocking(Socket sock) {
-#if defined(OPENSSL_WINDOWS)
-  u_long arg = 1;
-  return ioctlsocket(sock, FIONBIO, &arg) == 0;
-#else
-  int flags = fcntl(sock, F_GETFL, 0);
-  if (flags < 0) {
-    return false;
-  }
-  flags |= O_NONBLOCK;
-  return fcntl(sock, F_SETFL, flags) == 0;
-#endif
-}
-
-enum class WaitType { kRead, kWrite };
-
-static bool WaitForSocket(Socket sock, WaitType wait_type) {
-  // Use an arbitrary 5 second timeout, so the test doesn't hang indefinitely if
-  // there's an issue.
-  static const int kTimeoutSeconds = 5;
-#if defined(OPENSSL_WINDOWS)
-  fd_set read_set, write_set;
-  FD_ZERO(&read_set);
-  FD_ZERO(&write_set);
-  fd_set *wait_set = wait_type == WaitType::kRead ? &read_set : &write_set;
-  FD_SET(sock, wait_set);
-  timeval timeout;
-  timeout.tv_sec = kTimeoutSeconds;
-  timeout.tv_usec = 0;
-  if (select(0 /* unused on Windows */, &read_set, &write_set, nullptr,
-             &timeout) <= 0) {
-    return false;
-  }
-  return FD_ISSET(sock, wait_set);
-#else
-  short events = wait_type == WaitType::kRead ? POLLIN : POLLOUT;
-  pollfd fd = {/*fd=*/sock, events, /*revents=*/0};
-  return poll(&fd, 1, kTimeoutSeconds * 1000) == 1 && (fd.revents & events);
-#endif
-}
-
-TEST(BIOTest, SocketConnect) {
-  static const char kTestMessage[] = "test";
-  OwnedSocket listening_sock = ListenLoopback(/*backlog=*/1);
-  ASSERT_TRUE(listening_sock.is_valid()) << LastSocketError();
-
-  SockaddrStorage addr;
-  ASSERT_EQ(getsockname(listening_sock.get(), addr.addr_mut(), &addr.len), 0)
-      << LastSocketError();
-
-  char hostname[80];
-  if (addr.family() == AF_INET6) {
-    snprintf(hostname, sizeof(hostname), "[::1]:%d",
-             ntohs(addr.ToIPv6().sin6_port));
-  } else {
-    snprintf(hostname, sizeof(hostname), "127.0.0.1:%d",
-             ntohs(addr.ToIPv4().sin_port));
-  }
-
-  // Connect to it with a connect BIO.
-  bssl::UniquePtr<BIO> bio(BIO_new_connect(hostname));
-  ASSERT_TRUE(bio);
-
-  // Write a test message to the BIO. This is assumed to be smaller than the
-  // transport buffer.
-  ASSERT_EQ(static_cast<int>(sizeof(kTestMessage)),
-            BIO_write(bio.get(), kTestMessage, sizeof(kTestMessage)))
-      << LastSocketError();
-
-  // Accept the socket.
-  OwnedSocket sock(accept(listening_sock.get(), addr.addr_mut(), &addr.len));
-  ASSERT_TRUE(sock.is_valid()) << LastSocketError();
-
-  // Check the same message is read back out.
-  char buf[sizeof(kTestMessage)];
-  ASSERT_EQ(static_cast<int>(sizeof(kTestMessage)),
-            recv(sock.get(), buf, sizeof(buf), 0))
-      << LastSocketError();
-  EXPECT_EQ(Bytes(kTestMessage, sizeof(kTestMessage)), Bytes(buf, sizeof(buf)));
-}
-
-TEST(BIOTest, SocketNonBlocking) {
-  OwnedSocket listening_sock = ListenLoopback(/*backlog=*/1);
-  ASSERT_TRUE(listening_sock.is_valid()) << LastSocketError();
-
-  // Connect to |listening_sock|.
-  SockaddrStorage addr;
-  ASSERT_EQ(getsockname(listening_sock.get(), addr.addr_mut(), &addr.len), 0)
-      << LastSocketError();
-  OwnedSocket connect_sock(socket(addr.family(), SOCK_STREAM, 0));
-  ASSERT_TRUE(connect_sock.is_valid()) << LastSocketError();
-  ASSERT_EQ(connect(connect_sock.get(), addr.addr(), addr.len), 0)
-      << LastSocketError();
-  ASSERT_TRUE(SocketSetNonBlocking(connect_sock.get())) << LastSocketError();
-  bssl::UniquePtr<BIO> connect_bio(
-      BIO_new_socket(connect_sock.get(), BIO_NOCLOSE));
-  ASSERT_TRUE(connect_bio);
-
-  // Make a corresponding accepting socket.
-  OwnedSocket accept_sock(
-      accept(listening_sock.get(), addr.addr_mut(), &addr.len));
-  ASSERT_TRUE(accept_sock.is_valid()) << LastSocketError();
-  ASSERT_TRUE(SocketSetNonBlocking(accept_sock.get())) << LastSocketError();
-  bssl::UniquePtr<BIO> accept_bio(
-      BIO_new_socket(accept_sock.get(), BIO_NOCLOSE));
-  ASSERT_TRUE(accept_bio);
-
-  // Exchange data through the socket.
-  static const char kTestMessage[] = "hello, world";
-
-  // Reading from |accept_bio| should not block.
-  char buf[sizeof(kTestMessage)];
-  int ret = BIO_read(accept_bio.get(), buf, sizeof(buf));
-  EXPECT_EQ(ret, -1);
-  EXPECT_TRUE(BIO_should_read(accept_bio.get())) << LastSocketError();
-
-  // Writing to |connect_bio| should eventually overflow the transport buffers
-  // and also give a retryable error.
-  int bytes_written = 0;
-  for (;;) {
-    ret = BIO_write(connect_bio.get(), kTestMessage, sizeof(kTestMessage));
-    if (ret <= 0) {
-      EXPECT_EQ(ret, -1);
-      EXPECT_TRUE(BIO_should_write(connect_bio.get())) << LastSocketError();
-      break;
-    }
-    bytes_written += ret;
-  }
-  EXPECT_GT(bytes_written, 0);
-
-  // |accept_bio| should readable. Drain it. Note data is not always available
-  // from loopback immediately, notably on macOS, so wait for the socket first.
-  int bytes_read = 0;
-  while (bytes_read < bytes_written) {
-    ASSERT_TRUE(WaitForSocket(accept_sock.get(), WaitType::kRead))
-        << LastSocketError();
-    ret = BIO_read(accept_bio.get(), buf, sizeof(buf));
-    ASSERT_GT(ret, 0);
-    bytes_read += ret;
-  }
-
-  // |connect_bio| should become writeable again.
-  ASSERT_TRUE(WaitForSocket(accept_sock.get(), WaitType::kWrite))
-      << LastSocketError();
-  ret = BIO_write(connect_bio.get(), kTestMessage, sizeof(kTestMessage));
-  EXPECT_EQ(static_cast<int>(sizeof(kTestMessage)), ret);
-
-  ASSERT_TRUE(WaitForSocket(accept_sock.get(), WaitType::kRead))
-      << LastSocketError();
-  ret = BIO_read(accept_bio.get(), buf, sizeof(buf));
-  EXPECT_EQ(static_cast<int>(sizeof(kTestMessage)), ret);
-  EXPECT_EQ(Bytes(buf), Bytes(kTestMessage));
-
-  // Close one socket. We should get an EOF out the other.
-  connect_bio.reset();
-  connect_sock.reset();
-
-  ASSERT_TRUE(WaitForSocket(accept_sock.get(), WaitType::kRead))
-      << LastSocketError();
-  ret = BIO_read(accept_bio.get(), buf, sizeof(buf));
-  EXPECT_EQ(ret, 0) << LastSocketError();
-  EXPECT_FALSE(BIO_should_read(accept_bio.get()));
-}
-
 TEST(BIOTest, Printf) {
   // Test a short output, a very long one, and various sizes around
   // 256 (the size of the buffer) to ensure edge cases are correct.
@@ -832,32 +549,113 @@ TEST(BIOTest, ExternalData) {
 }
 
 // Test that, on Windows, |BIO_read_filename| opens files in binary mode.
-TEST(BIOTest, BinaryMode) {
+TEST(BIOTest, FileMode) {
   if (SkipTempFileTests()) {
     GTEST_SKIP();
   }
 
-  TemporaryFile file;
-  ASSERT_TRUE(file.Init("\r\n"));
+  TemporaryFile temp;
+  ASSERT_TRUE(temp.Init("hello\r\nworld"));
 
-  // Reading from the file should give back the exact bytes we put in.
+  auto expect_file_contents = [](BIO *bio, const std::string &str) {
+    // Read more than expected, to make sure we've reached the end of the file.
+    std::vector<char> buf(str.size() + 100);
+    int len = BIO_read(bio, buf.data(), static_cast<int>(buf.size()));
+    ASSERT_GT(len, 0);
+    EXPECT_EQ(Bytes(buf.data(), len), Bytes(str));
+  };
+  auto expect_binary_mode = [&](BIO *bio) {
+    expect_file_contents(bio, "hello\r\nworld");
+  };
+  auto expect_text_mode = [&](BIO *bio) {
+#if defined(OPENSSL_WINDOWS)
+    expect_file_contents(bio, "hello\nworld");
+#else
+    expect_file_contents(bio, "hello\r\nworld");
+#endif
+  };
+
+  // |BIO_read_filename| should open in binary mode.
   bssl::UniquePtr<BIO> bio(BIO_new(BIO_s_file()));
   ASSERT_TRUE(bio);
-  ASSERT_TRUE(BIO_read_filename(bio.get(), file.path().c_str()));
-  char buf[2];
-  ASSERT_EQ(2, BIO_read(bio.get(), buf, 2));
-  EXPECT_EQ(Bytes(buf, 2), Bytes("\r\n"));
+  ASSERT_TRUE(BIO_read_filename(bio.get(), temp.path().c_str()));
+  expect_binary_mode(bio.get());
+  // |BIO_new_file| does not set |BIO_FP_TEXT|, so expect it to set the file's
+  // mode to binary in all cases. This odd behavior, but it's what OpenSSL does.
+  bio.reset(BIO_new_file(temp.path().c_str(), "rb"));
+  ASSERT_TRUE(bio);
+  expect_binary_mode(bio.get());
+  bio.reset(BIO_new_file(temp.path().c_str(), "r"));
+  ASSERT_TRUE(bio);
+  // NOTE: Our behavior here aligns with OpenSSL which is to |_setmode| the file
+  // to binary. BoringSSL would |expect_text_mode| below because it respects 
+  // default mode on Windows which is text and doesn't call |_setmode| (unless
+  // |BIO_FP_TEXT| is set, which is not the case here).
+  expect_binary_mode(bio.get());
+  // |BIO_new_fp| will always set |_O_BINARY| if |BIO_FP_TEXT| is not set in the
+  // call to |BIO_new_fp|.
+  ScopedFILE file = temp.Open("rb");
+  ASSERT_TRUE(file);
+  bio.reset(BIO_new_fp(file.get(), BIO_NOCLOSE));
+  ASSERT_TRUE(bio);
+  expect_binary_mode(bio.get());
+  file = temp.Open("r");
+  ASSERT_TRUE(file);
+  bio.reset(BIO_new_fp(file.get(), BIO_NOCLOSE));
+  ASSERT_TRUE(bio);
+  // NOTE: Our behavior here aligns with OpenSSL. BoringSSL would
+  // |expect_text_mode| below because they don't call |_setmode| unless
+  // |BIO_FP_TEXT| is set.
+  expect_binary_mode(bio.get());
+  // However, |BIO_FP_TEXT| changes the file to be text mode, no matter how it
+  // was opened.
+  file = temp.Open("rb");
+  ASSERT_TRUE(file);
+  bio.reset(BIO_new_fp(file.get(), BIO_NOCLOSE | BIO_FP_TEXT));
+  ASSERT_TRUE(bio);
+  expect_text_mode(bio.get());
+  file = temp.Open("r");
+  ASSERT_TRUE(file);
+  bio.reset(BIO_new_fp(file.get(), BIO_NOCLOSE | BIO_FP_TEXT));
+  ASSERT_TRUE(bio);
+  expect_text_mode(bio.get());
+  // |BIO_new_fd| inherits the FD's existing mode.
+  ScopedFD fd = temp.OpenFD(kOpenReadOnlyBinary);
+  ASSERT_TRUE(fd.is_valid());
+  bio.reset(BIO_new_fd(fd.get(), BIO_NOCLOSE));
+  ASSERT_TRUE(bio);
+  expect_binary_mode(bio.get());
+  fd = temp.OpenFD(kOpenReadOnlyText);
+  ASSERT_TRUE(fd.is_valid());
+  bio.reset(BIO_new_fd(fd.get(), BIO_NOCLOSE));
+  ASSERT_TRUE(bio);
+  expect_text_mode(bio.get());
 }
 
-// Run through the tests twice, swapping |bio1| and |bio2|, for symmetry.
-class BIOPairTest : public testing::TestWithParam<bool> {};
+// BIOPairTest: A parameterized test fixture for testing BIO pair operations
+// This test class runs each test case with four different combinations:
+//   1. Normal order (bio1->bio2) with legacy callback
+//   2. Normal order (bio1->bio2) with extended callback
+//   3. Swapped order (bio2->bio1) with legacy callback
+//   4. Swapped order (bio2->bio1) with extended callback
+//
+// Parameters:
+//   std::tuple<bool, bool>:
+//     - First bool (get<0>): Controls BIO swapping
+//       * false = normal order (bio1->bio2)
+//       * true = swapped order (bio2->bio1)
+//     - Second bool (get<1>): Controls callback type
+//       * false = legacy callback (BIO_set_callback)
+//       * true = extended callback (BIO_set_callback_ex)
+class BIOPairTest : public testing::TestWithParam<std::tuple<bool, bool>> {};
 
 TEST_P(BIOPairTest, TestPair) {
   BIO *bio1, *bio2;
   ASSERT_TRUE(BIO_new_bio_pair(&bio1, 10, &bio2, 10));
   bssl::UniquePtr<BIO> free_bio1(bio1), free_bio2(bio2);
 
-  if (GetParam()) {
+  const bool should_swap = std::get<0>(GetParam());
+  if (should_swap) {
     std::swap(bio1, bio2);
   }
 
@@ -975,6 +773,21 @@ static long bio_cb_ex(BIO *b, int oper, const char *argp, size_t len, int argi,
   return ret;
 }
 
+static long bio_cb(BIO *b, int oper, const char *argp, int argi,
+                      long argl, long ret) {
+  if (test_count_ex >= CB_TEST_COUNT) {
+    return CALL_BACK_FAILURE;
+  }
+  param_b_ex[test_count_ex] = b;
+  param_oper_ex[test_count_ex] = oper;
+  param_argp_ex[test_count_ex] = argp;
+  param_argi_ex[test_count_ex] = argi;
+  param_argl_ex[test_count_ex] = argl;
+  param_ret_ex[test_count_ex] = ret;
+  test_count_ex++;
+  return ret;
+}
+
 static void bio_callback_cleanup() {
   // These mocks are used in multiple tests and need to be reset
   test_count_ex = 0;
@@ -996,11 +809,24 @@ TEST_P(BIOPairTest, TestCallbacks) {
   BIO *bio1, *bio2;
   ASSERT_TRUE(BIO_new_bio_pair(&bio1, 10, &bio2, 10));
 
-  if (GetParam()) {
+  // Check the second parameter from the test tuple to determine which callback type to use
+  // params is a tuple<bool, bool> where:
+  //   - get<0> controls bio swapping
+  //   - get<1> controls callback type (true = extended, false = legacy)
+  const auto& params = GetParam();
+  if (std::get<0>(params)) {  // swap_bios
     std::swap(bio1, bio2);
   }
 
-  BIO_set_callback_ex(bio2, bio_cb_ex);
+  if (std::get<1>(params)) {
+    // Use extended callback (BIO_callback_ex) which provides additional parameters:
+    // - len: size of the buffer for read/write operations
+    // - processed: pointer to store number of bytes actually processed
+    BIO_set_callback_ex(bio2, bio_cb_ex);
+  } else {
+    // Use legacy callback (BIO_callback) with basic parameters
+    BIO_set_callback(bio2, bio_cb);
+  }
 
   // Data written in one end may be read out the other.
   uint8_t buf[TEST_BUF_LEN];
@@ -1020,25 +846,36 @@ TEST_P(BIOPairTest, TestCallbacks) {
   // The calls before the BIO operation use 1 for the BIO's return value
   ASSERT_EQ(param_ret_ex[0], 1);
 
-  // The calls after the BIO call use the return value from the BIO, which is
-  // the length of data read/written
-  ASSERT_EQ(param_ret_ex[1], TEST_DATA_WRITTEN);
 
-  // For callback_ex the |len| param is the requested number of bytes to
-  // read/write
-  ASSERT_EQ(param_len_ex[0], (size_t)TEST_BUF_LEN);
-  ASSERT_EQ(param_len_ex[0], (size_t)TEST_BUF_LEN);
 
-  // For callback_ex argi and arl are unused
-  ASSERT_EQ(param_argi_ex[0], 0);
-  ASSERT_EQ(param_argi_ex[1], 0);
-  ASSERT_EQ(param_argl_ex[0], 0);
-  ASSERT_EQ(param_argl_ex[1], 0);
+  if (!std::get<1>(params)) {
+    // Legacy callback uses ret to hold data processed
+    ASSERT_EQ(param_ret_ex[1], TEST_DATA_WRITTEN);
+    // Legacy callback argi is used for len
+    ASSERT_EQ(param_argi_ex[0], TEST_BUF_LEN);
+    ASSERT_EQ(param_argi_ex[1], TEST_BUF_LEN);
+    ASSERT_EQ(param_argl_ex[0], 0);
+    ASSERT_EQ(param_argl_ex[1], 0);
+  }
 
-  // processed is null (0 in the array) the first call and the actual data the
-  // second time
-  ASSERT_EQ(param_processed_ex[0], 0u);
-  ASSERT_EQ(param_processed_ex[1], 5u);
+  // Extended callback specific verifications
+  if (std::get<1>(params)) {// If using extended callback
+    // For callback_ex ret is set to 1 after setting processed
+    ASSERT_EQ(param_ret_ex[1], 1);
+    // For callback_ex argi and arl are unused
+    ASSERT_EQ(param_argi_ex[0], 0);
+    ASSERT_EQ(param_argi_ex[1], 0);
+    ASSERT_EQ(param_argl_ex[0], 0);
+    ASSERT_EQ(param_argl_ex[1], 0);
+    // For callback_ex the |len| param is the requested number of bytes to
+    // read/write
+    ASSERT_EQ(param_len_ex[0], (size_t)TEST_BUF_LEN);
+    ASSERT_EQ(param_len_ex[0], (size_t)TEST_BUF_LEN);
+    // processed is null (0 in the array) the first call and the actual data the
+    // second time
+    ASSERT_EQ(param_processed_ex[0], 0u);
+    ASSERT_EQ(param_processed_ex[1], 5u);
+  }
 
   // The mock should be "full" at this point
   ASSERT_EQ(test_count_ex, CB_TEST_COUNT);
@@ -1089,7 +926,33 @@ TEST(BIOTest, InvokeConnectCallback) {
 }
 }  // namespace
 
-INSTANTIATE_TEST_SUITE_P(All, BIOPairTest, testing::Values(false, true));
+// Instantiate the parameterized test suite for BIOPairTest
+// This creates test instances for all combinations of the boolean parameters
+//
+// Parameters:
+//   First argument "All": Test suite name prefix
+//   Second argument "BIOPairTest": The test fixture class
+//   Third argument: Parameter generator using testing::Combine to create Cartesian product
+//     - First Bool(): Controls BIO swapping
+//       * false: Keep original BIO order (bio1->bio2)
+//       * true:  Swap BIOs (bio2->bio1)
+//     - Second Bool(): Controls callback type
+//       * false: Use legacy callback (BIO_set_callback)
+//       * true:  Use extended callback (BIO_set_callback_ex)
+//
+// This will generate four test combinations:
+//   1. (false, false) - Normal order, Legacy callback
+//   2. (false, true)  - Normal order, Extended callback
+//   3. (true, false)  - Swapped order, Legacy callback
+//   4. (true, true)   - Swapped order, Extended callback
+INSTANTIATE_TEST_SUITE_P(
+All,
+  BIOPairTest,
+  testing::Combine(
+      testing::Bool(),  // swap_bios
+      testing::Bool()   // use_extended_callback
+  )
+  );
 
 TEST(BIOTest, ReadWriteEx) {
   bssl::UniquePtr<BIO> bio(BIO_new(BIO_s_mem()));
@@ -1206,88 +1069,145 @@ TEST(BIOTest, TestPutsNullMethod) {
 }
 } //namespace
 
-TEST(BIOTest, TestPutsCallbacks) {
+TEST_P(BIOPairTest, TestPutsCallbacks) {
   bio_callback_cleanup();
-  BIO* bio = BIO_new(BIO_s_mem());
+  bssl::UniquePtr<BIO> bio(BIO_new(BIO_s_mem()));
   ASSERT_TRUE(bio);
 
-  BIO_set_callback_ex(bio, bio_cb_ex);
+  //   - get<1> controls callback type (true = extended, false = legacy)
+  const bool use_extended_callback = std::get<1>(GetParam());
 
-  EXPECT_EQ(TEST_DATA_WRITTEN, BIO_puts(bio, "12345"));
+  if (use_extended_callback) {
+    // Use extended callback (BIO_callback_ex) which provides additional parameters:
+    // - len: size of the buffer for read/write operations
+    // - processed: pointer to store number of bytes actually processed
+    BIO_set_callback_ex(bio.get(), bio_cb_ex);
+  } else {
+    // Use legacy callback (BIO_callback) with basic parameters
+    BIO_set_callback(bio.get(), bio_cb);
+  }
+
+  EXPECT_EQ(TEST_DATA_WRITTEN, BIO_puts(bio.get(), "12345"));
 
   ASSERT_EQ(param_oper_ex[0], BIO_CB_PUTS);
   ASSERT_EQ(param_oper_ex[1], BIO_CB_PUTS | BIO_CB_RETURN);
 
   ASSERT_EQ(param_argp_ex[0], param_argp_ex[1]);
   ASSERT_EQ(param_ret_ex[0], 1);
-  ASSERT_EQ(param_ret_ex[1], TEST_DATA_WRITTEN);
 
-  // len unused in puts callback
-  ASSERT_FALSE(param_len_ex[0]);
-  ASSERT_FALSE(param_len_ex[1]);
+  if (!use_extended_callback) {
+    // Legacy callback uses ret for data processed
+    ASSERT_EQ(param_ret_ex[1], TEST_DATA_WRITTEN);
+  }
+
 
   ASSERT_EQ(param_argi_ex[0], 0);
   ASSERT_EQ(param_argi_ex[1], 0);
   ASSERT_EQ(param_argl_ex[0], 0);
   ASSERT_EQ(param_argl_ex[1], 0);
 
-  ASSERT_EQ(param_processed_ex[0], 0u);
-  ASSERT_EQ(param_processed_ex[1], 5u);
+  // Extended callback specific verifications
+  if (use_extended_callback) {
+    // ret is set to processed and ret is reset back to 1
+    ASSERT_EQ(param_ret_ex[1], 1);
+    // len unused in puts callback
+    ASSERT_FALSE(param_len_ex[0]);
+    ASSERT_FALSE(param_len_ex[1]);
+
+    // Verify processed bytes
+    ASSERT_EQ(param_processed_ex[0], 0u);
+    ASSERT_EQ(param_processed_ex[1], 5u);
+  }
 
   // The mock should be "full" at this point
   ASSERT_EQ(test_count_ex, CB_TEST_COUNT);
 
   bio_callback_cleanup();
-  ASSERT_EQ(BIO_free(bio), 1);
 }
 
-TEST(BIOTest, TestGetsCallback) {
+TEST_P(BIOPairTest, TestGetsCallback) {
   bio_callback_cleanup();
 
-  BIO* bio = BIO_new(BIO_s_mem());
+  bssl::UniquePtr<BIO> bio (BIO_new(BIO_s_mem()));
   ASSERT_TRUE(bio);
   // write data to BIO, then set callback
-  EXPECT_EQ(TEST_DATA_WRITTEN, BIO_write(bio, "12345", TEST_DATA_WRITTEN));
+  EXPECT_EQ(TEST_DATA_WRITTEN, BIO_write(bio.get(), "12345", TEST_DATA_WRITTEN));
   char buf[TEST_BUF_LEN];
-  BIO_set_callback_ex(bio, bio_cb_ex);
 
-  ASSERT_EQ(TEST_DATA_WRITTEN, BIO_gets(bio, buf, sizeof(buf)));
+  //   - get<1> controls callback type (true = extended, false = legacy)
+  const bool use_extended_callback = std::get<1>(GetParam());
+
+  if (use_extended_callback) {
+    // Use extended callback (BIO_callback_ex) which provides additional parameters:
+    // - len: size of the buffer for read/write operations
+    // - processed: pointer to store number of bytes actually processed
+    BIO_set_callback_ex(bio.get(), bio_cb_ex);
+  } else {
+    // Use legacy callback (BIO_callback) with basic parameters
+    BIO_set_callback(bio.get(), bio_cb);
+  }
+
+  ASSERT_EQ(TEST_DATA_WRITTEN, BIO_gets(bio.get(), buf, sizeof(buf)));
 
   ASSERT_EQ(param_oper_ex[0], BIO_CB_GETS);
   ASSERT_EQ(param_oper_ex[1], BIO_CB_GETS | BIO_CB_RETURN);
 
   ASSERT_EQ(param_argp_ex[0], param_argp_ex[1]);
   ASSERT_EQ(param_ret_ex[0], 1);
-  ASSERT_EQ(param_ret_ex[1], TEST_DATA_WRITTEN);
 
-  ASSERT_EQ(param_len_ex[0], (size_t)TEST_BUF_LEN);
-  ASSERT_EQ(param_len_ex[1], (size_t)TEST_BUF_LEN);
 
-  ASSERT_EQ(param_argi_ex[0], 0);
-  ASSERT_EQ(param_argi_ex[1], 0);
   ASSERT_EQ(param_argl_ex[0], 0);
   ASSERT_EQ(param_argl_ex[1], 0);
 
-  ASSERT_EQ(param_processed_ex[0], 0u);
-  ASSERT_EQ(param_processed_ex[1], 5u);
+  // Old-style callback uses argi for len
+  if (!use_extended_callback) {
+    // use ret for data processed
+    ASSERT_EQ(param_ret_ex[1], TEST_DATA_WRITTEN);
+    ASSERT_EQ(param_argi_ex[0], TEST_BUF_LEN);
+    ASSERT_EQ(param_argi_ex[1], TEST_BUF_LEN);
+  }
+
+  // Extended callback specific verifications
+  if (use_extended_callback) {
+    // ret is set to 1 after setting processed
+    ASSERT_EQ(param_ret_ex[1], 1);
+    ASSERT_EQ(param_argi_ex[0], 0);
+    ASSERT_EQ(param_argi_ex[1], 0);
+    ASSERT_EQ(param_len_ex[0], (size_t)TEST_BUF_LEN);
+    ASSERT_EQ(param_len_ex[1], (size_t)TEST_BUF_LEN);
+    ASSERT_EQ(param_processed_ex[0], 0u);
+    ASSERT_EQ(param_processed_ex[1], 5u);
+  }
+
 
   ASSERT_EQ(test_count_ex, CB_TEST_COUNT);
 
   bio_callback_cleanup();
-  ASSERT_EQ(BIO_free(bio), 1);
 }
 
-TEST(BIOTest, TestCtrlCallback) {
+TEST_P(BIOPairTest, TestCtrlCallback) {
   bio_callback_cleanup();
 
-  BIO* bio = BIO_new(BIO_s_mem());
+  bssl::UniquePtr<BIO> bio(BIO_new(BIO_s_mem()));
   ASSERT_TRUE(bio);
-  BIO_set_callback_ex(bio, bio_cb_ex);
+
+  //   - get<1> controls callback type (true = extended, false = legacy)
+  const bool use_extended_callback = std::get<1>(GetParam());
+
+  if (use_extended_callback) {
+    // Use extended callback (BIO_callback_ex) which provides additional parameters:
+    // - len: size of the buffer for read/write operations
+    // - processed: pointer to store number of bytes actually processed
+    BIO_set_callback_ex(bio.get(), bio_cb_ex);
+  } else {
+    // Use legacy callback (BIO_callback) with basic parameters
+    BIO_set_callback(bio.get(), bio_cb);
+  }
 
   char buf[TEST_BUF_LEN];
   // Test BIO_ctrl. This is not normally called directly so we can use one of
   // the macros such as BIO_reset to test it
-  ASSERT_EQ(BIO_reset(bio), 1);
+  ASSERT_EQ(BIO_reset(bio.get()), 1);
 
   ASSERT_EQ(param_oper_ex[0], BIO_CB_CTRL);
   ASSERT_EQ(param_oper_ex[1], BIO_CB_CTRL | BIO_CB_RETURN);
@@ -1306,8 +1226,24 @@ TEST(BIOTest, TestCtrlCallback) {
 
   bio_callback_cleanup();
   // check that BIO_reset was called correctly
-  ASSERT_EQ(BIO_gets(bio, buf, sizeof(buf)), 0);
+  ASSERT_EQ(BIO_gets(bio.get(), buf, sizeof(buf)), 0);
 
   bio_callback_cleanup();
-  ASSERT_EQ(BIO_free(bio), 1);
+}
+
+TEST(BIOTest, GetMemDataBackwardsCompat) {
+  bssl::UniquePtr<BIO> bio(BIO_new(BIO_s_mem()));
+  ASSERT_TRUE(bio);
+
+  const uint8_t contents[] = {0x72, 0x61, 0x63, 0x63, 0x6f, 0x6f, 0x6e};
+
+  // Write some test data
+  int write_len = BIO_write(bio.get(), contents, sizeof(contents));
+  ASSERT_EQ(sizeof(contents), (size_t)write_len);
+
+  // Yes, this is something gRPC does
+  const uint8_t *ptr = NULL;
+  long data_len = BIO_get_mem_data(bio.get(), &ptr);
+  ASSERT_EQ((size_t)data_len, sizeof(contents));
+  EXPECT_EQ(Bytes(contents, sizeof(contents)), Bytes(ptr, data_len));
 }
